import { ConflictException, Injectable } from '@nestjs/common'
import { Game, MoneyUpdate, MoneyUpdateType } from '@prisma/client'
import { PrismaService } from 'src/prisma/prisma.service'
import { RepairGame } from './dto/repair-game.dto'

@Injectable()
export class GameService {
  constructor (private prisma: PrismaService) {}

  async find(): Promise<Game> {
    return await this.prisma.game.findUnique({ where: { id: 1 } })
  }

  async start(): Promise<Game> {
    const game = await this.find()
    if (game.started)
      throw new ConflictException("La partie a déjà démarré.")
    const players = await this.prisma.player.findMany()
    const alreadyStarted = game.currentRunnerId !== null
    if (!alreadyStarted) {
      for (const player of players) {
        await this.prisma.moneyUpdate.create({
          data: {
            playerId: player.id,
            amount: 2000,
            reason: MoneyUpdateType.START,
          }
        })
      }
    }
    const runnerId = alreadyStarted ? game.currentRunnerId : players[Math.floor(players.length * Math.random())].id
    return await this.prisma.game.update({
      where: { id: 1 },
      data: {
        started: true,
        currentRunnerId: runnerId,
      },
    })
  }

  async switchRunningPlayer(): Promise<Game> {
    const game = await this.find()
    const newRunnerId = game.currentRunnerId == 1 ? 2 : 1
    await this.prisma.moneyUpdate.create({
      data: {
        playerId: newRunnerId,
        amount: 300,
        reason: MoneyUpdateType.START,
      }
    })
    return await this.prisma.game.update({
      where: { id: game.id },
      data: { currentRunnerId: newRunnerId },
    })
  }

  async stop(): Promise<Game> {
    const game = await this.find()
    if (!game.started)
      throw new ConflictException("La partie n'a pas encore démarré.")
    return await this.prisma.game.update({
      where: { id: 1 },
      data: {
        started: false,
      },
    })
  }

  async reset(): Promise<Game> {
    await this.prisma.moneyUpdate.deleteMany()
    await this.prisma.challengeAction.deleteMany()
    await this.prisma.trainTrip.deleteMany()
    await this.prisma.geolocation.deleteMany()
    await this.prisma.player.updateMany({ data: { money: 0 } })
    await this.prisma.game.update({
      where: { id: 1 },
      data: {
        started: false,
        currentRunnerId: null,
      },
    })
    return await this.find()
  }

  async repair(): Promise<RepairGame> {
    const added: MoneyUpdate[] = []
    const deleted: MoneyUpdate[] = []
    const trains = await this.prisma.trainTrip.findMany({ include: { moneyUpdate: true } })
    for (const train of trains) {
      if (!train.moneyUpdate) {
        const trainMoneyUpdate = await this.prisma.moneyUpdate.create({
          data: {
            playerId: train.playerId,
            amount: -10 * Math.ceil(train.distance / 1000),
            reason: MoneyUpdateType.BUY_TRAIN,
            tripId: train.id,
          }
        })
        added.push(trainMoneyUpdate)
      }
    }
    const orpanTrainMoneyUpdates = await this.prisma.moneyUpdate.findMany({ where: { reason: MoneyUpdateType.BUY_TRAIN, tripId: null } })
    await this.prisma.moneyUpdate.deleteMany({ where: { reason: MoneyUpdateType.BUY_TRAIN, tripId: null } })
    deleted.push(...orpanTrainMoneyUpdates)

    const challengeActions = await this.prisma.challengeAction.findMany({ include: { moneyUpdate: true, challenge: true } })
    for (const challengeAction of challengeActions) {
      if (challengeAction.success && !challengeAction.moneyUpdate) {
        const challengeMoneyUpdate = await this.prisma.moneyUpdate.create({
          data: {
            playerId: challengeAction.playerId,
            amount: challengeAction.challenge.reward,
            reason: MoneyUpdateType.WIN_CHALLENGE,
            actionId: challengeAction.id,
          }
        })
        added.push(challengeMoneyUpdate)
      }
      else if (!challengeAction.success && challengeAction.moneyUpdate) {
        deleted.push(challengeAction.moneyUpdate)
        await this.prisma.moneyUpdate.delete({ where: { id: challengeAction.moneyUpdate.id } })
      }
    }

    const orpanChallengeMoneyUpdates = await this.prisma.moneyUpdate.findMany({ where: { reason: MoneyUpdateType.WIN_CHALLENGE, actionId: null } })
    await this.prisma.moneyUpdate.deleteMany({ where: { reason: MoneyUpdateType.WIN_CHALLENGE, actionId: null } })
    deleted.push(...orpanTrainMoneyUpdates)

    return { added: added, deleted: deleted }
  }
}
